import random

import pytest

from cocodev.django.deployment_configuration.models import Option


@pytest.mark.django_db
def test_retrieval(deployment_settings_module, django_assert_num_queries):
    ds = deployment_settings_module

    # Assure ourselves that we have an uncontaminated environment
    from cocodev.django.deployment_configuration import magic
    assert magic.MAGICKED_MODULE is None, "Deployment settings wasn't replaced"

    # Fully load the library, replacing the module in place
    from cocodev.django.deployment_configuration.load import fully_load_library
    fully_load_library()
    assert magic.MAGICKED_MODULE == ds, "Deployment settings wasn't replaced"

    # Create a random value to test with (this can be improved with hypothesis)
    expected = random.randint(-1000, 1000)
    for x in ('VAR_A', 'VAR_B'):
        var = Option.objects.filter(name=x).first()
        var.raw_value = str(expected)
        var.save()

    with django_assert_num_queries(0):
        # No queries are made because the values have already been loaded into the cache

        # Check that the random value was retrieved
        assert ds.VAR_B == expected, "VAR_B was not retrieved correctly"
        assert ds.VAR_A == expected, "VAR_A was not retrieved correctly"



from cocodev.django.deployment_configuration.core import load_deployment_settings_module
from cocodev.django.deployment_configuration.django_settings_helpers import (
    LIBRARY_SETTINGS,
)


def module_from_library_settings():
    return load_deployment_settings_module(
        LIBRARY_SETTINGS.deployment_settings_file
    )



"""
    Dummy conftest.py for deployment_configuration.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    - https://docs.pytest.org/en/stable/fixture.html
    - https://docs.pytest.org/en/stable/writing_plugins.html
"""

from importlib import reload

import pytest

from cocodev.django.deployment_configuration import magic
from cocodev.django.deployment_configuration.shortcuts import (
    module_from_library_settings,
)


@pytest.fixture()
def deployment_settings_module():
    """
    Convenience fixture to access the deployment settings module in the
    example project
    """
    import tests.example_project.example_project.deployment_settings
    return tests.example_project.example_project.deployment_settings


@pytest.fixture(autouse=True)
def refresh_deployment_settings_module():
    """
    If a test calls `fully_load_library` then the deployment settings file will
    always be magicked, which can cause cross-contamination with other test cases.
    This can be alleviated with the use of `pytest-random-order`, but it can still
    fly under the radar long enough for the faulty commit to be hidden in the history.
    """
    deployment_settings_module = module_from_library_settings()
    yield
    reload(deployment_settings_module)
    magic.MAGICKED_MODULE = None



from django import forms
from django.contrib import admin
from django.contrib.admin.templatetags.admin_list import _boolean_icon
from django.core.exceptions import ValidationError
from django.utils.safestring import mark_safe

from cocodev.django.deployment_configuration import deployment_settings
from cocodev.django.deployment_configuration.core import (
    generate_type_string_from_type,
    hydrate_value,
)
from cocodev.django.deployment_configuration.django_settings_helpers import (
    LIBRARY_SETTINGS,
)
from cocodev.django.deployment_configuration.models import Option, OptionType
from cocodev.django.deployment_configuration.store import change_value_for_option


class OptionForm(forms.ModelForm):
    class Meta:
        model = Option
        exclude = []

    class Media:
        js = ["deployment_configuration/admin_live_update.js"]

    def __init__(self, *args, instance=None, **kwargs):
        super().__init__(*args, instance=instance, **kwargs)

        if instance:
            self.fields['option_type'].queryset = OptionType.objects.filter(supportedoptions=instance)
            self.fields['raw_value'].required = False

    def clean(self):
        data = super().clean()
        if 'raw_value' not in data:
            data['raw_value'] = False

        raw_value = data['raw_value']
        option_type = data['option_type']

        try:
            hydrate_value(
                raw_value,
                option_type.python_callable,
            )
            change_value_for_option(
                self.instance,
                raw_value,
                option_type,
            )
        except Exception as e:
            raise ValidationError(str(e)) from e

        return data


@admin.register(Option)
class OptionAdmin(admin.ModelAdmin):
    fieldsets = [
        (
            "Basic Information",
            {
                "fields": ["name", "current_value", "current_cache_value", "matches_cache", ],
            },
        ),
        (
            "Make Changes",
            {
                "fields": ["raw_value", "option_type", ]
            }
        ),
        (
            "Defaults",
            {
                "classes": ["collapse"],
                "fields": ["default_type_with_label", "default_value", ],
            }
        ),
        (
            "Documentation",
            {
                "classes": ["collapse", "start-open", ],
                "fields": ["rendered_documentation", ],
            },
        ),
    ]
    if not LIBRARY_SETTINGS.use_cache:
        fieldsets[0][1]['fields'] = list(filter(lambda v: 'cache' not in v, fieldsets[0][1]['fields']))

    readonly_fields = [
        'name', 'current_value', 'current_cache_value',
        'default_type_with_label', 'default_value',
        'rendered_documentation', 'matches_cache'
    ]
    if not LIBRARY_SETTINGS.use_cache:
        readonly_fields = list(filter(lambda v: 'cache' not in v, readonly_fields))

    exclude = ['supported_types', 'documentation', 'default_type', 'allowed_types', ]
    list_display = [
        'updated',
        'current_value', 'current_cache_value', 'matches_cache',
        'default_value', 'default_type',
    ]
    if not LIBRARY_SETTINGS.use_cache:
        list_display = list(filter(lambda v: 'cache' not in v, list_display))

    actions = ["reset_to_default", "reset_cache_value", ]
    if not LIBRARY_SETTINGS.use_cache:
        actions = list(filter(lambda v: 'cache' not in v, actions))

    form = OptionForm

    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).select_related('option_type', 'default_type')

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def formfield_for_dbfield(self, db_field, request, **kwargs):
        formfield = super().formfield_for_dbfield(db_field, request, **kwargs)

        match db_field.name:
            case "option_type":
                formfield.empty_label = None
                formfield.widget.can_delete_related = False
                formfield.widget.can_change_related = False
                formfield.widget.can_add_related = False
                formfield.widget.can_view_related = False
                formfield.label_from_instance = lambda obj: obj.admin_label()

        return formfield

    @admin.display(description="Description")
    def rendered_documentation(self, instance):
        return mark_safe(instance.documentation)

    @admin.display(description="Default type")
    def default_type_with_label(self, instance):
        return instance.default_type.admin_label()

    @admin.display(description="Current value")
    def current_value(self, obj: Option):
        if obj.option_type.python_callable == 'builtins.bool':
            value = _boolean_icon(obj.value)
        else:
            value = obj.value

        return mark_safe(f"{obj.option_type.python_callable}: {value}")

    @admin.display(description="Current cached value")
    def current_cache_value(self, obj: Option):
        value = getattr(deployment_settings, obj.name)
        type_string = generate_type_string_from_type(type(value))

        return mark_safe(f"{type_string}: {value}")

    @admin.display(description="Cache matches")
    def matches_cache(self, obj: Option):
        return _boolean_icon(self.current_cache_value(obj) == self.current_value(obj))

    @admin.action(description="Reset to default")
    def reset_to_default(self, request, queryset):
        option: Option
        for option in queryset:
            option.reset_to_default()
            option.save()

    @admin.action(description="Reset cache to match current value")
    def reset_cache_value(self, request, queryset):
        option: Option
        for option in queryset:
            option.save()


@admin.register(OptionType)
class OptionTypeAdmin(admin.ModelAdmin):
    readonly_fields = ['python_callable', 'documentation']

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False



import logging
from typing import Type

from django.db.models.signals import post_save
from django.dispatch import receiver

from cocodev.django.deployment_configuration.django_settings_helpers import (
    LIBRARY_SETTINGS,
)
from cocodev.django.deployment_configuration.models import Option
from cocodev.django.deployment_configuration.retrieve import get_option_from_db

log = logging.getLogger(__name__)


@receiver(post_save, sender=Option, )
def surgical_cache_clear(
        sender: Type[Option],
        instance: Option,
        created: bool,
        raw: bool,
        using: str,
        update_fields,
        **kwargs,
):
    """
    Clear the cache for the deployment setting that was just updated. Immediately
    after put the new value in the cache. Doing this will avoid hitting the database
    in any code which uses deployment settings minus scenarios with a TTL.
    """
    if LIBRARY_SETTINGS.use_cache:
        log.info("Clearing cache for %s", instance.name)
        get_option_from_db.clear_cache(
            key=get_option_from_db.make_key(instance.name)
        )
        # We could avoid hitting the database again here, but we'd have to make sure we
        # exactly replicate the cache function logic. In favor of being DRY we take a slight
        # performance hit.
        get_option_from_db(instance.name)



# Generated by Django 4.2.9 on 2024-04-15 22:46

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='OptionType',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('python_callable', models.CharField(max_length=2048)),
                ('documentation', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='Option',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=2048)),
                ('value', models.CharField(max_length=2048)),
                ('default_value', models.CharField(max_length=2048)),
                ('default_value_change_behavior', models.CharField(choices=[('always', 'Always'), ('never', 'Never'), ('change_if_unchanged', 'Change If Unchanged')], max_length=2048)),
                ('documentation', models.TextField()),
                ('default_type', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='defaultoptiontype', to='deployment_configuration.optiontype')),
                ('option_type', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='deployment_configuration.optiontype')),
                ('supported_types', models.ManyToManyField(related_name='supportedoptiontypes', to='deployment_configuration.optiontype')),
            ],
        ),
    ]



from typing import Annotated

from cocodev.django.deployment_configuration.data import Metadata
from cocodev.django.deployment_configuration.typing import DefaultChangeBehavior

VAR_A: Annotated[
    int | str | set,  # Allowed types
    Metadata(
        documentation="This is a <em>basic example</em> of deployment configuration",
        behavior_when_default_changes=DefaultChangeBehavior.NEVER,
    )
] = 99  # Default value and type
VAR_B: float = 99.12321



from django.db import models
from django.utils.translation import gettext_lazy as _


class DefaultChangeBehavior(models.TextChoices):
    ALWAYS = 'always_change', _("Always Change")
    NEVER = 'never_change', _("Never Change")
    CHANGE_IF_UNCHANGED = 'change_if_unchanged', _("Change if Unchanged")



#!/usr/bin/env python
"""Since we are trying to distribute Django applications as Python Packages,
it is important that some commands (such `migrate`) are available from within
the installation of the package without the need of copying the source code.

Because of that ``pyscaffoldext-django`` moves the generated ``__main__.py`` file
to become the package's ``__main__.py`` file. This way all the commands that
could be run before as ``python3 __main__.py <COMMAND>`` can now be run as
``python3 -m cocodev.services.community_creator <COMMAND>``, in a straight forward fashion just after
a ``pip3 install``.

This file is a executable stub that simply calls ``__main__.py:main()`` for
the sake of backward compatibility of the developer's workflow.
"""
import os
import sys

# This makes the package usable even without being installed with pip
# (redundant in the case the developer uses `pip install -e .`)
sys.path.insert(0, os.path.join(os.path.dirname(os.path.abspath(__file__)), 'src'))


if __name__ == '__main__':
    from tests.example_project.manage import main

    main()



import importlib
import importlib.util


def callable_from_string(callable_string):
    module_str, attr_name = callable_string.rsplit('.', maxsplit=1)
    module = importlib.import_module(module_str)
    return getattr(module, attr_name)


def import_from_filepath(filepath):
    """
    Copied from https://stackoverflow.com/a/67692

    Dynamically load a filepath as a python module and insert it into the
    system modules.

    Args:
        filepath: A string or filepath to load as a python module

    Returns: The python module represented by the input string
    """
    import importlib.util
    import sys
    module_name = "cocodev.django.deployment_configuration.tmp.deployment_settings"
    spec = importlib.util.spec_from_file_location(
        module_name,
        filepath,
    )
    foo = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = foo
    spec.loader.exec_module(foo)

    return foo



import pytest

from cocodev.django.deployment_configuration.models import Option, OptionType


@pytest.mark.django_db
def test_value_property_hydrates_properly():
    db_default_type = OptionType.objects.create(python_callable='builtins.list', documentation='Just for this test')
    option = Option.objects.create(
        name="UNIT_TEST",
        raw_value="[{9, 1, '55'}, {'r': 't'}, 88, 'test']",
        option_type=db_default_type,
        default_type=db_default_type,
        default_value='{0, 0}',
        default_value_change_behavior="always_change",
        documentation="Unit testing",
    )

    actual = option.value
    expected = [{9, 1, '55'}, {'r': 't'}, 88, 'test']

    assert actual == expected



import pytest

from cocodev.django.deployment_configuration.store import push_data_to_database


@pytest.mark.django_db
def test_push_data_to_database_works_on_happy_path():
    option = push_data_to_database(
        option_name="TEST_VAR",
        default_value="Hello, World!",
        default_type="builtins.str",
        supported_types=["builtins.str", "builtins.float"],
        default_behavior="always_change",
        help_string="This is a test value",
    )

    assert option.name == 'TEST_VAR'
    assert option.raw_value == option.default_value == 'Hello, World!'
    assert option.option_type.python_callable == option.default_type.python_callable == 'builtins.str'
    assert list(option.supported_types.all().values_list('python_callable', flat=True)) == ["builtins.str", "builtins.float"]
    assert option.default_value_change_behavior == 'always_change'
    assert option.documentation == "This is a test value"



# Generated by Django 5.0.6 on 2024-08-05 18:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('deployment_configuration', '0003_alter_option_supported_types'),
    ]

    operations = [
        migrations.AlterField(
            model_name='option',
            name='default_value_change_behavior',
            field=models.CharField(choices=[('always_change', 'Always Change'), ('never_change', 'Never Change'), ('change_if_unchanged', 'Change if Unchanged')], default='always_change', max_length=2048),
        ),
    ]



from django.apps import AppConfig
from django.db.models.signals import post_migrate

from cocodev.django.deployment_configuration.exceptions import MissingTable
from cocodev.django.deployment_configuration.load import fully_load_library


def fully_load_library_after_migrations(sender, **kwargs):
    fully_load_library()


class DeploymentConfigurationConfig(AppConfig):
    default = True
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'cocodev.django.deployment_configuration'

    def ready(self):
        try:
            fully_load_library()
        except MissingTable:
            """
            If migrations are running the first call will fail, but the
            post_migrate signal is only sent when migrations are running,
            so we can't use that normally.
            """
            post_migrate.connect(
                fully_load_library_after_migrations,
                sender=self,
            )


class DeploymentConfigurationTestConfig(AppConfig):
    """
    This exists just to not autoload settings during tests, which
    can make tests more difficult than necessary.
    """
    default = False
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'cocodev.django.deployment_configuration'



import logging

from cocodev.django.deployment_configuration.retrieve import (
    get_all_option_names,
    get_option_from_db,
)

log = logging.getLogger(__name__)

MAGICKED_MODULE = None


def replace_deployment_settings_module(module):
    for option in get_all_option_names():
        delattr(module, option)

    module.__getattr__ = get_option_from_db

    global MAGICKED_MODULE
    if MAGICKED_MODULE and MAGICKED_MODULE != module:
        log.error("Applying module magic twice may result in strange behavior and bugs!")

    MAGICKED_MODULE = module



from tests.example_project.example_project.settings import *  # no-qa

try:
    INSTALLED_APPS.remove(
        'cocodev.django.deployment_configuration',
    )
except ValueError:
    ...

test_mod = 'cocodev.django.deployment_configuration.apps.DeploymentConfigurationTestConfig'
if test_mod not in INSTALLED_APPS:
    INSTALLED_APPS.insert(
        -1,
        test_mod,
    )



__all__ = ["async_deployment_settings"]

from typing import Coroutine

from cocodev.django.deployment_configuration.shortcuts import (
    module_from_library_settings,
)


class _AsyncDeploymentSettings:
    def __getattr__(self, item) -> Coroutine:
        return self.get_deployment_setting(item)

    def get_deployment_setting(self, name):
        return getattr(module_from_library_settings(), name)


async_deployment_settings = _AsyncDeploymentSettings()



from cocodev.django.deployment_configuration.core import load_deployment_settings_module


def test_load_deployment_settings_module_with_path(settings):
    from tests.example_project.example_project import deployment_settings

    result = load_deployment_settings_module(settings.BASE_DIR / "example_project" / 'deployment_settings.py')
    expected = deployment_settings

    assert result.__spec__.origin == expected.__spec__.origin, "Did not load the settings file correctly"


def test_load_deployment_settings_module_with_path_string(settings):
    from tests.example_project.example_project import deployment_settings

    string = str(settings.BASE_DIR / "example_project" / 'deployment_settings.py')
    result = load_deployment_settings_module(string)
    expected = deployment_settings

    assert result.__spec__.origin == expected.__spec__.origin, "Did not load the settings file correctly"


def test_load_deployment_settings_module_with_module_path():
    from tests.example_project.example_project import deployment_settings

    string = 'tests.example_project.example_project.deployment_settings'
    result = load_deployment_settings_module(string)
    expected = deployment_settings

    assert result.__spec__.origin == expected.__spec__.origin, "Did not load the settings file correctly"



import asyncio
import random

import pytest

from cocodev.django.deployment_configuration.load import fully_load_library
from cocodev.django.deployment_configuration.models import Option
from cocodev.django.deployment_configuration.shortcuts import (
    module_from_library_settings,
)


@pytest.mark.django_db
def test_some_asyncio_code():
    fully_load_library()

    # Create a random value to test with (this can be improved with hypothesis)
    expected = random.randint(-1000, 1000)
    for x in ('VAR_A', 'VAR_B'):
        var = Option.objects.filter(name=x).first()
        var.raw_value = str(expected)
        var.save()

    did_run = False

    # Show that access from an async context works as long as caching is set up correctly
    async def testing():
        nonlocal did_run
        ds = module_from_library_settings()
        assert ds.VAR_B == expected
        did_run = True

    asyncio.run(testing())

    # Just to make sure this test doesn't pass because of a developer error
    assert did_run



from django.core.exceptions import ImproperlyConfigured


class BaseDjangoDeploymentConfigurationException(Exception):
    """
    Convenience for catching all library errors.
    """
    ...


class InvalidTypeForOption(BaseDjangoDeploymentConfigurationException):
    """
    An Option was saved with a type that it does not support.
    """
    ...


class LibraryIsImproperlyConfigured(ImproperlyConfigured, BaseDjangoDeploymentConfigurationException):
    """
    Oopsies, you didn't follow the instructions!
    """


class MissingOptionInDatabase(BaseDjangoDeploymentConfigurationException):
    """
    Did you load the library before accessing the deployment options?
    """


class MissingTable(BaseDjangoDeploymentConfigurationException):
    """
    Migrations need to be run!
    """



import logging

from django.db import OperationalError

from cocodev.django.deployment_configuration.exceptions import MissingTable
from cocodev.django.deployment_configuration.shortcuts import (
    module_from_library_settings,
)

log = logging.getLogger(__name__)
SHOULD_LOAD = True


def fully_load_library():
    """
    Fully loads the library and slides the deployment
    Returns:
    """
    from cocodev.django.deployment_configuration.core import (
        generate_deployment_settings_safely,
    )
    from cocodev.django.deployment_configuration.magic import (
        replace_deployment_settings_module,
    )
    from cocodev.django.deployment_configuration.metadata import _APP

    # Load all of our signals
    __import__(f"cocodev.django.deployment_configuration.signals")

    deployment_settings_module = module_from_library_settings()
    generate_deployment_settings_safely(
        deployment_settings_module
    )
    if _APP['loaded'] == 'module':
        try:
            replace_deployment_settings_module(deployment_settings_module)
        except OperationalError as e:
            from cocodev.django.deployment_configuration.models import Option
            if (
                    f"no such table: {Option._meta.db_table}"
                    in repr(e)
            ):
                raise MissingTable("Migrations need to be run!") from e



# Generated by Django 4.2.9 on 2024-04-15 23:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('deployment_configuration', '0001_initial'),
    ]

    operations = [
        migrations.RenameField(
            model_name='option',
            old_name='value',
            new_name='raw_value',
        ),
    ]



# Generated by Django 4.2.9 on 2024-04-17 20:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('deployment_configuration', '0002_rename_value_option_raw_value'),
    ]

    operations = [
        migrations.AlterField(
            model_name='option',
            name='supported_types',
            field=models.ManyToManyField(related_name='supportedoptions', to='deployment_configuration.optiontype'),
        ),
    ]



import pytest

from cocodev.django.deployment_configuration.core import (
    generate_deployment_settings_safely,
)
from cocodev.django.deployment_configuration.models import Option


@pytest.mark.django_db
def test_generate_deployment_settings(deployment_settings_module):
    generate_deployment_settings_safely(deployment_settings_module)
    option = Option.objects.filter(name='VAR_A').first()

    assert option.name == 'VAR_A'
    assert option.raw_value == option.default_value == '99'
    assert option.option_type.python_callable == option.default_type.python_callable == 'builtins.int'
    assert set(
        option.supported_types.all().values_list('python_callable', flat=True)
    ) == {"builtins.str", "builtins.int", "builtins.set"}
    assert option.default_value_change_behavior == 'never_change'
    assert option.documentation == "This is a <em>basic example</em> of deployment configuration"



"""
Where the fun stuff happens
"""
import ast
import importlib
import logging
import typing
from functools import lru_cache, singledispatch
from pathlib import Path
from types import ModuleType

import django.db.utils

from cocodev.django.deployment_configuration.data import Metadata
from cocodev.django.deployment_configuration.import_helper import (
    callable_from_string,
    import_from_filepath,
)
from cocodev.django.deployment_configuration.metadata import _APP

log = logging.getLogger(__name__)

HYDRATION_MAP: typing.Annotated[
    dict[str, typing.Callable],
    """
    This is required to make the builtin functions work off string values. Maybe
    there is another way to sneakily slide these callables over top of the base
    callable, but this works well enough. For your own callables, make sure that they
    can accept a string and return a value.
    """
] = {
    'builtins.list': lambda v: list(ast.literal_eval(v)),
    'builtins.set': lambda v: set() if v == "{}" else set(ast.literal_eval(v)) ,
    'builtins.dict': lambda v: dict(ast.literal_eval(v)),
    'builtins.NoneType': lambda v: None,
}


@lru_cache(maxsize=1)
def load_deployment_settings_module(file_or_module_path: Path | str, /) -> ModuleType:
    """
    Loads the deployment settings module from a string or Path. This
    is how we transform the settings value into the actual python it
    refers to.

    Args:
        file_or_module_path: A string, filepath, or module path to load
            as a python module

    Returns: The python module represented by the input string
    """
    try:
        if isinstance(file_or_module_path, Path) or Path(file_or_module_path).exists():
            _APP['loaded'] = 'file'
            return import_from_filepath(file_or_module_path)

    except TypeError:
        ...

    _APP['loaded'] = 'module'
    return importlib.import_module(file_or_module_path)


def generate_type_string_from_variable(variable) -> str:
    """
    Accepts a variable returns an importable module string
    """
    the_type = type(variable)
    return generate_type_string_from_type(the_type)


def find_metadata(variable_type) -> typing.Optional[Metadata]:
    """
    Searches through the annotated metadata for our Metadata class
    """
    if hasattr(variable_type, "__metadata__"):
        for value in variable_type.__metadata__:
            if isinstance(value, Metadata):
                return value


def generate_type_string_from_type(the_type):
    """
    Generates a callable string from a type.

    Example:
        >>> assert generate_type_string_from_type(str) == "builtins.str"

    Args:
        the_type: Any type that can be used within this library

    Returns: A callable string for use within this library
    """
    type_name = the_type.__qualname__
    module = the_type.__module__
    return f"{module}.{type_name}"


def generate_deployment_settings_safely(deployment_settings_module: ModuleType):
    try:
        generate_deployment_settings(deployment_settings_module)
    except django.db.utils.OperationalError:
        # Probably migrations
        log.error(
            "Could not create deployment settings in database! "
            "If you're running migrations you can safely ignore this."
        )


def generate_deployment_settings(deployment_settings_module: ModuleType):
    """
    Iterates over the deployment settings file and generates the necessary
    data in the database.
    """
    # This function runs once, so we pay for the local import instead of paying
    # every time a value is accessed
    from cocodev.django.deployment_configuration.store import (
        clean_up_old_options,
        push_data_to_database,
    )

    for variable_name, variable_type in deployment_settings_module.__annotations__.items():
        default_value = getattr(deployment_settings_module, variable_name)
        default_type = generate_type_string_from_variable(default_value)
        metadata = find_metadata(variable_type) or Metadata()

        if type_args := typing.get_args(variable_type):
            raw_supported_types = type_args[0]
            supported_types = {
                generate_type_string_from_type(supported_type)
                for supported_type
                in typing.get_args(raw_supported_types)
            }
        else:
            supported_types = {
                generate_type_string_from_type(variable_type)
            }

        push_data_to_database(
            option_name=variable_name,
            default_value=default_value,
            default_type=default_type,
            supported_types=supported_types,
            default_behavior=metadata.behavior_when_default_changes,
            help_string=metadata.documentation,
        )

    all_deployment_option_names = deployment_settings_module.__annotations__.keys()
    clean_up_old_options(all_deployment_option_names)


def hydrate_value(value, callable_str):
    """
    Converts a value and a callable string into the value returned
    by the callable when passed the value.

    TODO allow registering a function for a callable to allow for customization

    Args:
        value: The input to the callable represented by the string
        callable_str: A string which represents a callable (ex: builtins.str)

    Returns: The converted value of the input
    """
    actual_callable = HYDRATION_MAP.get(
        callable_str,
        callable_from_string(callable_str),
    )
    return actual_callable(value)


@singledispatch
def dehydrate_value(new_value: typing.Any) -> str:
    """
    Serializes the value into a string. Register a function with a type to
    customize the serialization.
    """
    return str(new_value)



import dataclasses

from cocodev.django.deployment_configuration.typing import DefaultChangeBehavior


@dataclasses.dataclass(frozen=True, kw_only=True)
class Metadata:
    documentation: str = ''
    behavior_when_default_changes: DefaultChangeBehavior = DefaultChangeBehavior.ALWAYS


Metadata.DefaultChangeBehavior = DefaultChangeBehavior



__all__ = [
    "get_all_option_names",
    "get_option_from_db",
]

from cocodev.django.deployment_configuration.django_cache_helpers import ttl_cache
from cocodev.django.deployment_configuration.django_settings_helpers import (
    LIBRARY_SETTINGS,
)
from cocodev.django.deployment_configuration.exceptions import (
    LibraryIsImproperlyConfigured,
    MissingOptionInDatabase,
)
from cocodev.django.deployment_configuration.models import Option


def get_all_option_names():
    return set(Option.objects.all().values_list('name', flat=True))


def get_option_from_db(key):
    if key not in get_all_option_names():
        raise AttributeError(
            f"Deployment settings "
            f"'{LIBRARY_SETTINGS.deployment_settings_file}'"
            f" has no attribute '{key}'"
        )

    option = Option.objects.filter(name=key).first()
    if option:
        return option.value
    else:
        raise MissingOptionInDatabase("Could not find option in database, is the library loaded?")


if LIBRARY_SETTINGS.use_cache:
    _cache_opts = {}
    for _opt_key, _key in {('ttl', 'cache_ttl'), ('cache_name', 'cache_name')}:
        try:
            _cache_opts[_opt_key] = getattr(LIBRARY_SETTINGS, _key)
        except LibraryIsImproperlyConfigured:
            ...

    get_option_from_db = ttl_cache(
        get_option_from_db,
        **_cache_opts,
    )



from cocodev.django.deployment_configuration.import_helper import import_from_filepath


def test_loads_file_correctly(settings):
    from tests.example_project.example_project import deployment_settings

    result = import_from_filepath(settings.BASE_DIR / "example_project" / 'deployment_settings.py')
    expected = deployment_settings

    assert result.__spec__.origin == expected.__spec__.origin, "Did not load the settings file correctly"



import functools
from functools import _make_key
from typing import Callable, Optional

from django.core.cache import caches


def ttl_cache(function: Callable = None, /, ttl: Optional[float] = None, cache_name: str = 'default'):
    """
    A TTL cache backed by Django's cache framework. See the CACHES setting
    for information on how the cache operates.

    https://docs.djangoproject.com/en/4.2/topics/cache/
    """

    def get_cache():
        return caches[cache_name]

    def clear_cache(key=None):
        """
        Perhaps allowing the clearing the entire cache is dangerous
        """
        if key:
            return get_cache().delete(key)
        else:
            return get_cache().clear()

    # https://stackoverflow.com/a/74650070/1112794
    def decorator(func):

        def make_key(*args, **kwargs):
            return str(
                _make_key((func.__module__, func.__qualname__, *args), kwargs, typed=False)
            ).replace(
                '[', '_bo_'
            ).replace(
                ']', '_bc_'
            ).replace(
                ' ',
                '',
            )

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            key = make_key(*args, **kwargs)
            results = caches[cache_name].get(key)
            if not results:
                results = func(*args, **kwargs)
                caches[cache_name].set(key, results, ttl)

            return results

        wrapper.clear_cache = clear_cache
        wrapper.get_cache = get_cache
        wrapper.make_key = make_key

        return wrapper

    if function:
        return decorator(function)
    else:
        return decorator



import logging
from typing import Iterable

from cocodev.django.deployment_configuration.exceptions import InvalidTypeForOption
from cocodev.django.deployment_configuration.import_helper import callable_from_string
from cocodev.django.deployment_configuration.models import Option, OptionType
from cocodev.django.deployment_configuration.typing import DefaultChangeBehavior

log = logging.getLogger(__name__)


def push_data_to_database(
        *,
        option_name: str,
        default_value: str,
        default_type: str,
        supported_types: Iterable[str],
        default_behavior: DefaultChangeBehavior,
        help_string: str,
):
    """
    Pushes the deployment option into the database and creates the related
    models.

    Args:
        option_name:
        default_value:
        default_type:
        supported_types:
        default_behavior:
        help_string:

    Returns:

    """
    # Validate some things
    if default_type not in supported_types:
        raise InvalidTypeForOption(
            f"Invalid type: {option_name} supports {supported_types} but "
            f"has a default of {default_type}"
        )

    db_default_type, _ = generate_option_type(default_type)
    db_supported_types = [generate_option_type(supported_type)[0] for supported_type in supported_types]

    option: Option = Option.objects.filter(name=option_name).first()
    if not option:
        # Option does not exist yet
        option = Option.objects.create(
            name=option_name,
            raw_value=default_value,
            option_type=db_default_type,
            default_type=db_default_type,
            default_value=default_value,
            default_value_change_behavior=default_behavior,
            documentation=help_string,
        )
        option.supported_types.add(*db_supported_types)
    else:
        # Option exists

        # Update the default behavior first
        option.default_value_change_behavior = default_behavior

        # If the value always changes with the default
        if option.default_value_change_behavior == Option.DefaultChangeBehavior.ALWAYS:
            option.raw_value = default_value
            option.option_type = db_default_type

        # If the value changes if it is still the old default
        if (
                option.default_value_change_behavior == Option.DefaultChangeBehavior.CHANGE_IF_UNCHANGED
                and option.raw_value == option.default_value
        ):
            option.raw_value = default_value
            option.option_type = db_default_type

        # Error correction
        if option.option_type.python_callable not in supported_types:
            log.error(
                "Configuration option's supported types were changed "
                "and do not include the existing type, resetting to default",
                extra={
                    'data': {
                        'option': option.name,
                        'old_value': option.raw_value,
                        'old_type': option.option_type,
                        'new_value': default_value,
                        'new_type': default_type,
                    }
                }
            )
            option.raw_value = default_value
            option.option_type = default_type

        option.default_value = default_value
        option.default_type = db_default_type
        option.documentation = help_string

        option.save()
        option.supported_types.add(*db_supported_types)

    return option


def generate_option_type(default_type) -> tuple[OptionType, bool]:
    return OptionType.objects.update_or_create(
        python_callable=default_type,
        defaults={
            'documentation': callable_from_string(default_type).__doc__
        }
    )


def clean_up_old_options(existing_options: Iterable[str]):
    old_options = Option.objects.exclude(name__in=existing_options)
    if old_options:
        log.warning("Deleting old options: %s", old_options.values_list('name', flat=True))

    return old_options.delete()


def change_value_for_option(option, raw_value, option_type: OptionType):
    if not option.supported_types.all().only('pk').filter(pk=option_type.pk).exists():
        raise ValueError(
            f"Cannot assign '{raw_value}' of type '{option_type.python_callable}' to {option.name}! "
            f"Supported types are {list(option.supported_types.all().values_list('python_callable', flat=True))}"
        )

    option.raw_value = raw_value
    option.option_type = option_type



from django.db import models

from cocodev.django.deployment_configuration.core import (
    dehydrate_value,
    generate_type_string_from_type,
    hydrate_value,
)
from cocodev.django.deployment_configuration.import_helper import callable_from_string
from cocodev.django.deployment_configuration.typing import DefaultChangeBehavior


def meta_fields() -> tuple[models.DateTimeField, models.DateTimeField]:
    """
    Convenience for creating consistent db metadata columns.
    """
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    return created_at, updated_at


class Option(models.Model):
    created, updated = meta_fields()

    DefaultChangeBehavior = DefaultChangeBehavior

    name = models.CharField(max_length=2048)
    raw_value = models.CharField(max_length=2048)
    option_type = models.ForeignKey(to='OptionType', on_delete=models.PROTECT)

    supported_types = models.ManyToManyField(to='OptionType', related_name='supportedoptions')

    default_value = models.CharField(max_length=2048)
    default_type = models.ForeignKey(
        to='OptionType',
        on_delete=models.PROTECT,
        related_name='defaultoptiontype',
    )
    default_value_change_behavior = models.CharField(
        max_length=2048,
        choices=DefaultChangeBehavior.choices,
        blank=False,
        default=DefaultChangeBehavior.ALWAYS,
    )

    documentation = models.TextField()

    @property
    def value(self):
        """
        Returns the hydrated value from the database.
        """
        return hydrate_value(self.raw_value, self.option_type.python_callable)

    @value.setter
    def value(self, new_value):
        from cocodev.django.deployment_configuration.store import generate_option_type

        new_option_type_str = generate_type_string_from_type(type(new_value))
        new_option_type = generate_option_type(new_option_type_str)

        if new_option_type.pk not in self.supported_types.all().only('pk'):
            raise ValueError(
                f"Cannot assign {new_value} of type {type(new_value)} to {self.name}! "
                f"Supported types are {self.supported_types.all().values_list('python_callable', flat=True)}"
            )

        self.raw_value = dehydrate_value(new_value)
        self.option_type = new_option_type

    def reset_to_default(self):
        self.raw_value = self.default_value
        self.option_type = self.default_type

    def __str__(self):
        return self.name


class OptionType(models.Model):
    created, updated = meta_fields()

    python_callable = models.CharField(max_length=2048)

    documentation = models.TextField()

    def admin_label(self):
        return self.python_callable

    def to_callable(self):
        return callable_from_string(self.python_callable)

    def __str__(self):
        return self.python_callable



from unittest.mock import sentinel

from django.conf import settings

from cocodev.django.deployment_configuration.exceptions import (
    LibraryIsImproperlyConfigured,
)

notset = sentinel.notset


def get_library_setting(key, default=notset):
    try:
        return settings.DEPLOYMENT_CONFIGURATION_SETTINGS[key]
    except:
        if default is not notset:
            return default
        else:
            raise LibraryIsImproperlyConfigured(
                f"You must set {key} in the DEPLOYMENT_CONFIGURATION_SETTINGS "
                f"settings dict for the library to function"
            )


class LibrarySettings:
    @property
    def use_cache(self):
        return get_library_setting('use_cache', True)

    @property
    def deployment_settings_file(self):
        return get_library_setting('deployment_settings_file')

    @property
    def cache_name(self):
        return get_library_setting('cache_name', 'default')

    @property
    def cache_ttl(self):
        return get_library_setting('cache_ttl', None)


LIBRARY_SETTINGS = LibrarySettings()
